/*
 * Copyright (c) 2004 bartavelle
 * bartavelle at bandecon.com
 */

#include <string.h>

#include "path.h"
#include "arch.h"
#include "misc.h"
#include "common.h"
#include "formats.h"
#include "sha.h"

#define FORMAT_LABEL			"raw-sha1-opencl"
#define FORMAT_NAME			"Raw SHA-1 OpenCL"
#define ALGORITHM_NAME			"raw-sha1-opencl"
#define SHA_TYPE                        "SHA-1"
#define BENCHMARK_COMMENT		""
#define BENCHMARK_LENGTH		0

#define PLAINTEXT_LENGTH		32
#define CIPHERTEXT_LENGTH		40

#define BINARY_SIZE			20
#define SALT_SIZE			0
#define SSHA_NUM_KEYS         		1

#define MIN_KEYS_PER_CRYPT		SSHA_NUM_KEYS
#define MAX_KEYS_PER_CRYPT		SSHA_NUM_KEYS

#ifndef uint32_t
#define uint32_t unsigned int
#endif

typedef struct {
    uint32_t h0,h1,h2,h3,h4;
} SHA_DEV_CTX;


cl_platform_id platform;
cl_device_id devices;
cl_context context;
cl_program program;
cl_command_queue queue;
cl_int ret_code;
cl_kernel ssha_crypt_kernel;
cl_mem pinned_saved_keys, pinned_partial_hashes, buffer_out, buffer_keys, buffer_hash, len_buffer;
static SHA_DEV_CTX *outbuffer;
static int lenpwd[SSHA_NUM_KEYS];
unsigned char *inbuffer;
static size_t global_work_size = SSHA_NUM_KEYS;
static size_t local_work_size = 1;

static struct fmt_tests rawsha1_tests[] = {
	{"a9993e364706816aba3e25717850c26c9cd0d89d", "abc"},
	{"2fbf0eba37de1d1d633bc1ed943b907f9b360d4c", "azertyuiop1"},
	{"f879f8090e92232ed07092ebed6dc6170457a21d", "azertyuiop2"},
	{"1813c12f25e64931f3833b26e999e26e81f9ad24", "azertyuiop3"},
	{NULL}
};

static char saved_key[PLAINTEXT_LENGTH];
static ARCH_WORD_32 crypt_key[BINARY_SIZE / 4];

static void if_error_log(cl_int ret_code, const char *message)
{
    if(ret_code != CL_SUCCESS) {
        printf("\nOpenCL: %s\n", message);
        exit(-1);
    }
}

static int valid(char *ciphertext)
{
	int i;

	if (strlen(ciphertext) != CIPHERTEXT_LENGTH) return 0;
	for (i = 0; i < CIPHERTEXT_LENGTH; i++){
		if (!(  (('0' <= ciphertext[i])&&(ciphertext[i] <= '9')) ||
					(('a' <= ciphertext[i])&&(ciphertext[i] <= 'f'))  
					|| (('A' <= ciphertext[i])&&(ciphertext[i] <= 'F'))))
			return 0;
	}
	return 1;
}

static void rawsha1_set_salt(void *salt) { }

static void rawsha1_opencl_init(void)
{
    // load kernel source
    char *source=(char*)mem_alloc(1024*16);
    //FILE *fp = fopen(path_expand("$JOHN/sha1_opencl_kernel.cl"),"r");
    FILE *fp = fopen(path_expand("$JOHN/ssha_opencl_kernel.cl"),"r");
    if(!fp)
        if_error_log(!CL_SUCCESS, "Source kernel not found!");
    size_t source_size = fread(source, sizeof(char), 1024*16, fp);
    source[source_size] = 0;
    fclose(fp);

    // get a platform and its information
    size_t max_group_size;
    //CHECK
    char log[1024*64];
    ret_code = clGetPlatformIDs(1, &platform, NULL);
    if_error_log(ret_code, "No OpenCL platform exist");
    ret_code = clGetPlatformInfo(platform, CL_PLATFORM_NAME, sizeof(log), log, NULL);
    if_error_log(ret_code, "Error querying PLATFORM_NAME");
    printf("\nOpenCL Platform: <<<%s>>>", log);

    // find an OpenCL device
    //ret_code = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &devices, NULL);
    ret_code = clGetDeviceIDs(platform, CL_DEVICE_TYPE_CPU, 1, &devices, NULL);
    if_error_log(ret_code, "No OpenCL device of that type exist");
    ret_code = clGetDeviceInfo(devices, CL_DEVICE_NAME, sizeof(log), log, NULL);
    if_error_log(ret_code, "Error querying DEVICE_NAME");
    printf(" and device: <<<%s>>>\n",log);
    ret_code = clGetDeviceInfo(devices, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(max_group_size), &max_group_size, NULL);
    if_error_log(ret_code, "Error querying MAX_WORK_GROUP_SIZE");

    // create a context and command queue on the device.
    context = clCreateContext(NULL, 1, &devices, NULL, NULL, &ret_code);
    //if_error_log(ret_code, "Error creating context");
        if (ret_code != CL_SUCCESS) {
			printf("errore num %d",ret_code);
			exit(2);
	}
    queue = clCreateCommandQueue(context, devices, 0, &ret_code);
    if_error_log(ret_code, "Error creating command queue");

    // submit the kernel source for compilation
    program = clCreateProgramWithSource(context, 1, (const char **)&source, NULL, &ret_code);
    if_error_log(ret_code,"Error creating program");
    ret_code = clBuildProgram(program, 1, &devices, NULL, NULL, NULL);
    if(ret_code != CL_SUCCESS) {
        printf("failed in clBuildProgram with %d\n", ret_code);
        clGetProgramBuildInfo(program, devices, CL_PROGRAM_BUILD_LOG, sizeof(log), (void*)log, NULL);
        printf("compilation log: %s\n", log);
        exit(-1);
    }

    // create kernel to execute
    ssha_crypt_kernel = clCreateKernel(program, "sha1_crypt_kernel", &ret_code);
    if_error_log(ret_code, "Error creating kernel. Double-check kernel name?");

    // create Page-Locked (Pinned) memory for higher bandwidth between host and device (Nvidia Best Practices)
    pinned_saved_keys = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR,
            (PLAINTEXT_LENGTH)*SSHA_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code, "Error creating page-locked memory");
    inbuffer = (unsigned char*)clEnqueueMapBuffer(queue, pinned_saved_keys, CL_TRUE, CL_MAP_WRITE | CL_MAP_READ, 0, (PLAINTEXT_LENGTH)*SSHA_NUM_KEYS, 0, NULL, NULL, &ret_code);
    if_error_log (ret_code, "Error mapping page-locked memory inbuffer");


    pinned_partial_hashes = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, sizeof(SHA_DEV_CTX)*SSHA_NUM_KEYS, NULL, &ret_code);

    if_error_log (ret_code, "Error creating page-locked memory");

    outbuffer = (SHA_DEV_CTX*)clEnqueueMapBuffer(queue, pinned_partial_hashes, CL_TRUE, CL_MAP_READ, 0, sizeof(SHA_DEV_CTX)*SSHA_NUM_KEYS, 0, NULL, NULL, &ret_code);
    if_error_log (ret_code, "Error mapping page-locked memory outbuffer");

    // create and set arguments
    buffer_keys = clCreateBuffer(context, CL_MEM_READ_ONLY, (PLAINTEXT_LENGTH)*SSHA_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code, "Error creating buffer keys argument");

    len_buffer = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(int)*SSHA_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code, "Error creating buffer len  argument");

    buffer_out  = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(SHA_DEV_CTX)*SSHA_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code,"Error creating buffer out argument");

    ret_code = clSetKernelArg(ssha_crypt_kernel, 0, sizeof(buffer_keys), (void*) &buffer_keys);
    if_error_log (ret_code, "Error setting argument 1");
    //ret_code = clSetKernelArg(ssha_crypt_kernel, 1, sizeof(buffer_out ), (void*) &buffer_out);
    //if_error_log (ret_code, "Error setting argument 2");
    ret_code = clSetKernelArg(ssha_crypt_kernel, 1, sizeof(len_buffer), (void*) &len_buffer);
    if_error_log (ret_code, "Error setting argument 3");
    ret_code = clSetKernelArg(ssha_crypt_kernel, 2, sizeof(buffer_out ), (void*) &buffer_out);
    if_error_log (ret_code, "Error setting argument 3");
}

static void rawsha1_set_key(char *key, int index) {
        memset(saved_key,0,PLAINTEXT_LENGTH);
	strnzcpy(saved_key, key, PLAINTEXT_LENGTH);
}

static char *rawsha1_get_key(int index) {
	return saved_key;
}

static int rawsha1_cmp_all(void *binary, int count) { 
	return !memcmp(binary, crypt_key, BINARY_SIZE);
}

static int rawsha1_cmp_exact(char *source, int count){
  return (1);
}

static int rawsha1_cmp_one(void * binary, int index)
{
	return rawsha1_cmp_all(binary, index);
}

static void rawsha1_crypt_all(int count) {  
        cl_int code;
	
	lenpwd[0] = strlen(saved_key);
	
	//saved_key[lenpwd[0]]='0x80';
	memcpy(inbuffer,saved_key,lenpwd[0]);
	inbuffer[lenpwd[0]] = 0x80;
    	code = clEnqueueWriteBuffer(queue, buffer_keys, CL_TRUE, 0, (PLAINTEXT_LENGTH) * SSHA_NUM_KEYS, inbuffer, 0, NULL, NULL);
    	if(code != CL_SUCCESS) {
        	printf("failed in clEnqueueWriteBuffer inbuffer with code %d\n", code);
        	exit(-1);
    	}
    	code = clEnqueueWriteBuffer(queue, len_buffer, CL_TRUE, 0, sizeof(int) * SSHA_NUM_KEYS, lenpwd, 0, NULL, NULL);

    	if(code != CL_SUCCESS) {
        	printf("failed in clEnqueueWriteBuffer p with code %d\n", code);
        	exit(-1);
	}
    	// execute ssha kernel
    	code = clEnqueueNDRangeKernel(queue, ssha_crypt_kernel, 1, NULL, &global_work_size, &local_work_size, 0, NULL, NULL);
    	if(code != CL_SUCCESS) {
        	printf("failed in clEnqueueNDRangeKernel with code %d\n", code);
       		exit(-1);
    	}
    	clFinish(queue);
    	// read back partial hashes
    	clEnqueueReadBuffer(queue, buffer_out, CL_TRUE, 0, sizeof(SHA_DEV_CTX)*SSHA_NUM_KEYS, outbuffer, 0, NULL, NULL);
	
	(*outbuffer).h0 = be32toh((*outbuffer).h0);
	(*outbuffer).h1 = be32toh((*outbuffer).h1);
	(*outbuffer).h2 = be32toh((*outbuffer).h2);
	(*outbuffer).h3 = be32toh((*outbuffer).h3);
	(*outbuffer).h4 = be32toh((*outbuffer).h4);
	memcpy(crypt_key,outbuffer,sizeof(SHA_DEV_CTX));
  
}

static void * rawsha1_binary(char *ciphertext) 
{
	static char realcipher[BINARY_SIZE];
	int i;
	
	for(i=0;i<BINARY_SIZE;i++)
	{
		realcipher[i] = atoi16[ARCH_INDEX(ciphertext[i*2])]*16 + atoi16[ARCH_INDEX(ciphertext[i*2+1])];
	}
	return (void *)realcipher;
}

static int binary_hash_0(void *binary)
{
	return ((ARCH_WORD_32 *)binary)[0] & 0xF;
}

static int binary_hash_1(void *binary)
{
	return ((ARCH_WORD_32 *)binary)[0] & 0xFF;
}

static int binary_hash_2(void *binary)
{
	return ((ARCH_WORD_32 *)binary)[0] & 0xFFF;
}

static int get_hash_0(int index)
{
	return ((ARCH_WORD_32 *)crypt_key)[index] & 0xF;
}

static int get_hash_1(int index)
{
	return ((ARCH_WORD_32 *)crypt_key)[index] & 0xFF;
}

static int get_hash_2(int index)
{
	return ((ARCH_WORD_32 *)crypt_key)[index] & 0xFFF;
}

struct fmt_main fmt_opencl_rawSHA1 = {
	{
		FORMAT_LABEL,
		FORMAT_NAME,
		SHA_TYPE,
		BENCHMARK_COMMENT,
		BENCHMARK_LENGTH,
		PLAINTEXT_LENGTH,
		BINARY_SIZE,
		SALT_SIZE,
		MIN_KEYS_PER_CRYPT,
		MAX_KEYS_PER_CRYPT,
		FMT_CASE | FMT_8_BIT,
		rawsha1_tests
	}, {
		rawsha1_opencl_init,
		valid,
		fmt_default_split,
		rawsha1_binary,
		fmt_default_salt,
		{
			binary_hash_0,
			binary_hash_1,
			binary_hash_2,
			NULL,
			NULL
		},
		fmt_default_salt_hash,
		rawsha1_set_salt,
		rawsha1_set_key,
		rawsha1_get_key,
		fmt_default_clear_keys,
		rawsha1_crypt_all,
		{
			get_hash_0,
			get_hash_1,
			get_hash_2,
			NULL,
			NULL
		},
		rawsha1_cmp_all,
		rawsha1_cmp_one,
		rawsha1_cmp_exact
	}
};
